#!/usr/bin/env python

""" quick view of detector images for a specified measurement

usage: irixs_oneshot [number of run]

"""

import sys
import os
import numpy as np
import matplotlib.pyplot as plt
import warnings

from glob import glob
from scipy.optimize import curve_fit
from mpl_toolkits.axes_grid1 import make_axes_locatable

from matplotlib.pyplot import imread

plt.rcParams['xtick.top'] = True
plt.rcParams['ytick.right'] = True
plt.rcParams['font.size'] = 8
plt.rcParams['axes.titlesize'] = 'medium'
plt.rcParams['figure.titlesize'] = 'medium'


def find_nearest(array, value):
    return (np.abs(array-value)).argmin()


def peak(x, a, sl, x0, f, bgnd):
    m = np.full(len(x), bgnd)
    sg = sl/np.sqrt(2*np.log(2))
    m += (1-f)*(a/(sg*np.sqrt(2.*np.pi))) * np.exp(-1.*(x-x0)**2./(2.*sg**2.))
    m += f * (a/np.pi) * (sl / ((x-x0)**2. + sl**2.))
    return m


def peak_fit(x, y):
    fra = 0.5
    bgnd = float(np.min(y))
    height = np.max(y)-bgnd
    half = x[find_nearest(y, height/2+bgnd)]
    cen = x[find_nearest(y, np.max(y))]
    fwhm = np.abs(half-cen)*2
    sig = fwhm / 2
    amp = height*(sig*np.sqrt(2.*np.pi))

    xf = np.linspace(x.min(), x.max(), 1000)
    p0 = [amp, sig, cen, fra, bgnd]
    bounds = [(0, 1E-6, -1E9, 0, 0),(1E9, 1E6, 1E9, 1, 1E9)]
    p, _ = curve_fit(peak, x, y, p0, bounds=bounds)
    yi = peak(xf, *p)

    return xf, yi, p


def load_fio(run, exp, datdir):
    path = '{0}/{1}_{2:05d}.fio'.format(datdir, exp, run)
    a = {}
    head = []
    data = []
    with open(path) as f:
        for line in f:
            l = line.strip()
            if line.startswith('%d'):
                break
        for line in f:
            l = line.strip().split()
            if not l:
                break
            if l[0] == 'Col':
                head.append(l[2])
            else:
                try:
                    data.append([float(x) for x in l])
                except ValueError:
                    break
    if head and data:
        data = [x for x in data if x]
        data = np.array(data)
        data = data.view(dtype=[(n, float) for n in head])
        data = data.reshape(len(data))
        a['data'] = data
        a['auto'] = head[0]
        a['pnts'] = data.shape[0]
        return a


def load_tiff(run, exp, datdir, no, indicator=True):
    path = '{0}/{1}_{2:05d}/andor/{1}_{2:05d}_{3:04d}.tiff'.format(
            datdir, exp, run, no)
    with warnings.catch_warnings():
        try:
            img = imread(path)
            if indicator:
                print('*', end=' ', flush=True)
        except (Warning, FileNotFoundError, OSError):
            print('\nload_tiff: failed to load {}'.format(path))
            return
    return img


def load(run, exp, datdir, detfac, to, co):
    a = load_fio(run, exp, datdir)
    imgtest = load_tiff(run, exp, datdir, 0, indicator=False)
    if imgtest is None:
        return
    else:
        a['img'] = []
    print('#{} ({} points)'.format(run, a['pnts']), end=' ')
    for i,_ in enumerate(a['data'][a['auto']]):
        if i > len(a['img'])-1:
            img = load_tiff(run, exp, datdir, i)
            if img is not None:
                img -= detfac
                img[~np.logical_and(img > to, img < co)] = 0
                a['img'].append(img)
    print()
    return a


def detector(run, exp, datdir, vmax=10, threshold=1010, cutoff=1800, detfac=935):

    to = threshold - detfac
    co = cutoff - detfac

    a = load(run, exp, datdir, detfac, to, co)
    if a is None:
        return

    step = a['auto']
    if step == 'exp_dmy01':
        oneshot = True
    else:
        oneshot = False

    imgarr = np.atleast_3d(np.array(a['img']))
    imtotal = np.nansum(imgarr, axis=0) / imgarr.shape[0]

    if oneshot:
        x = np.arange(imgarr.shape[1])
        y = np.nansum(imgarr, axis=(0,2)) / imgarr.shape[0]
    else:
        x = []
        y = []
        for i, ef in enumerate(a['data'][a['auto']]):
            try:
                img = imgarr[i]
            except IndexError:
                continue
            yi = np.nansum(img)
            xi = ef
            x.append(xi)
            y.append(yi)
        x, y = np.array(x), np.array(y)

    p = False
    if oneshot:  # try fitting the signal if the measurement was a oneshot
        try:
            xf, yf, p = peak_fit(x, y)
            print('  amp: {:.2f}'.format(p[0]), end='  ')
            print('fwhm: {:.3f}'.format(p[1]*2), end='  ')
            print('cen: {:.4f}'.format(p[2]), end='  ')
            print('fra: {:3.1f}'.format(p[3]), end='  ')
            print('bg: {:.2f}'.format(p[3]), end='')
        except:
            pass
    print()

    fig, ax = plt.subplots(1, 2, figsize=(8.5, 4))
    fig.subplots_adjust(0.06, 0.15, 0.98, 0.93)

    plt.suptitle(f'#{run}', ha='left', va='top', x=0.005, y=0.995)

    im = ax[0].imshow(imtotal, origin='lower', vmax=vmax,
                      cmap=plt.get_cmap('bone_r'),
                      interpolation='hanning')

    div = make_axes_locatable(ax[0])
    cax = div.append_axes('right', size='4%', pad=0.1)
    fig.colorbar(im, cax=cax)

    ax[0].set_title('Summed Detector Map')
    ax[0].set_xlabel('x-pixel')
    ax[0].set_ylabel('y-pixel')
    ax[0].tick_params(which='both', direction='out', length=2)
    ax[0].xaxis.set_major_locator(plt.MultipleLocator(400))
    ax[0].yaxis.set_major_locator(plt.MultipleLocator(400))
    ax[0].xaxis.set_minor_locator(plt.MultipleLocator(100))
    ax[0].yaxis.set_minor_locator(plt.MultipleLocator(100))

    # total counts
    ax[1].plot(x, y, lw=1, color='#001F3F')
    if oneshot:
        ax[1].set_xlabel('y-pixel')
        ax[1].set_title('Integrated')
    else:
        ax[1].ticklabel_format(axis='y', style='sci', scilimits=(0, 0))
        ax[1].set_title('Counts')

    if p is not False:
        ax[1].plot(xf, yf,
                   color='#F012BE', dashes=(2,8), lw=0.5)
        fr = 'fwhm: {:.3f}\ncen: {:.2f}'.format(
             p[1]*2,p[2])
        ax[1].text(0.025, 0.975, fr, va='top',
                   transform=ax[1].transAxes,
                   fontsize='small', linespacing=1.3)

    for axi in ax[1:]:
        axi.minorticks_on()
        if not oneshot:
            axi.set_xlabel(step)
            for l in axi.get_xmajorticklabels():
                l.set_rotation(30)


def find_run(run):
    datdir = '/gpfs/current/raw'
    search = datdir + '/*_{0:05d}.fio'.format(run)
    path = sorted(glob(search), key=os.path.getmtime)
    if not path:
        datdir = '/gpfs/commissioning/raw'
        search = datdir + '/*_{0:05d}.fio'.format(run)
        path = sorted(glob(search), key=os.path.getmtime)
    if path:
        path = path[-1]
        _,f = path.rsplit('/',1)
        exp,_ = f.rsplit('_',1)
        return exp, datdir
    else:
        return False, False


def main():
    try:
        run = int(sys.argv[1])
    except (IndexError, ValueError):
        print("irixs_oneshot [number of run]")
        sys.exit(2)
    exp, datdir = find_run(run)
    if exp:
        detector(run, exp, datdir)
        plt.show()
    else:
        print('failed to load #{}'.format(run))


if __name__ == "__main__":
    main()



""" Basic six-circle diffractometer simulator

References
----------
Uses methodology described by
H. You, 4S+2D six-circle diffractometer, J. Appl. Cryst. 32, 614, (1999)
https://doi.org/10.1107/S0021889899001223

Basic approach follows diffcalc (Diamond Light Source)
https://github.com/DiamondLightSource/diffcalc

See also:
W. R. Busing and H. A. Levy, Angle calculations for 3- and 4-circle X-ray and neutron
diffractometers, Acta Cryst. 22, 457-464, (1967) https://doi.org/10.1107/S0365110X67000970

https://github.com/mantidproject/documents/blob/master/Design/UBMatriximplementationnotes.pdf

Circle definitions
------------------
mu -- horiz th
nu -- horiz tth
eta -- vert th
delta -- vert tth
chi -- sample rocking
phi -- sample rotation

Notes
-----
UB matrix setup limited to IRIXS geometry (horizontal scattering angles)
th, tth, chi -> mu, nu, chi
"""

import numpy as np

from numpy import pi, sin, cos, radians, degrees
from numpy.linalg import norm, inv, multi_dot
from scipy.optimize import least_squares
from copy import deepcopy

from .tools import reciprocol_lattice

np.seterr(invalid='raise')  # catch failed UB-matrix due to parallel reflections

def rotation_matricies(mu, nu, chi, eta=0, delta=0, phi=0):
    """ matrices corresponding to the rotation of the circles """

    MU = np.array([
        [1, 0, 0],
        [0, cos(mu), -sin(mu)],
        [0, sin(mu), cos(mu)]
    ])

    NU = np.array([
        [1, 0, 0],
        [0, cos(nu), -sin(nu)],
        [0, sin(nu), cos(nu)]
    ])

    CHI = np.array([
        [cos(chi), 0, sin(chi)],
        [0, 1, 0],
        [-sin(chi), 0, cos(chi)]
    ])

    ETA = np.array([
        [cos(eta), sin(eta), 0],
        [-sin(eta), cos(eta), 0],
        [0, 0, 1]
    ])

    DELTA = np.array([
        [cos(delta), sin(delta), 0],
        [-sin(delta), cos(delta), 0],
        [0, 0, 1]
    ])

    PHI = np.array([
        [cos(phi), sin(phi), 0],
        [-sin(phi), cos(phi), 0],
        [0, 0, 1]
    ])

    return MU, NU, CHI, ETA, DELTA, PHI


def B_matrix(b1, b2, b3, beta2, beta3, a3, alpha1):
    """B-matrix: crystal lattice orientation"""
    B = np.array([
        [b1, b2 * cos(beta3), b3 * cos(beta2)],
        [0.0, b2 * sin(beta3), -b3 * sin(beta2) * cos(alpha1)],
        [0.0, 0.0, 2 * pi / a3]
    ])
    return B


def U_matrix(h1c, h2c, u1p, u2p):
    """U-matrix: instrument orientation"""
    t1c = h1c
    t3c = np.cross(h1c.ravel(), h2c.ravel()).reshape(3, 1)
    t2c = np.cross(t3c.ravel(), t1c.ravel()).reshape(3, 1)

    t1p = u1p
    t3p = np.cross(u1p.ravel(), u2p.ravel()).reshape(3, 1)
    t2p = np.cross(t3p.ravel(), t1p.ravel()).reshape(3, 1)

    try:
        Tc = np.hstack([t1c/norm(t1c), t2c/norm(t2c), t3c/norm(t3c)])
        Tp = np.hstack([t1p/norm(t1p), t2p/norm(t2p), t3p/norm(t3p)])
    except FloatingPointError:
        raise BaseException("Reflections are parallel.")

    return Tp.dot(inv(Tc))


def q_phi(mu, nu, chi, eta=0, delta=0, phi=0):
    """Calculate hkl in phi frame, in units of 2*pi/lambda"""

    MU, NU, CHI, ETA, DELTA, PHI = rotation_matricies(mu, nu, chi, eta, delta, phi)

    q_lab = NU.dot(DELTA) - np.eye(3)
    q_lab = q_lab.dot(np.array([[0], [1], [0]]))

    Z = multi_dot([inv(PHI), inv(CHI), inv(ETA), inv(MU), q_lab])
    return Z


def q_hkl(wl, UB, mu, nu, chi, eta=0, delta=0, phi=0):
    """Calculate miller indicies from six circle angles with UB-matrix"""

    MU, NU, CHI, ETA, DELTA, PHI = rotation_matricies(mu, nu, chi, eta, delta, phi)

    q_lab = NU.dot(DELTA) - np.eye(3)
    q_lab = q_lab.dot(np.array([[0], [2 * pi / wl], [0]]))

    hkl = multi_dot([inv(UB), inv(PHI), inv(CHI), inv(ETA), inv(MU), q_lab])
    return hkl.ravel()


class sixc:
    """ Six-circle Diffractometer Simulator

    Attributes
    ----------
    UB : (3x3) np.array
        The UB-matrix maps scattering geometry in the laboratory frame (U) to
        the reciprocol lattice of the sample (B)
    orientation : list
        Reflection values for U orientation matrix [hkl0, hkl1, angles0, angles1]
    cell : list
        Lattice parameters [a(Å), b(Å), c(Å), alpha(rad), beta(rad), gamma(rad)]
    recip_cell : list
        Reciprocol lattice [a*, b*, c*, alpha*, beta*, gamma*]
    energy : float
        X-ray energy (eV)
    wl : float
        X-ray wavelength (Å)

    Methods
    -------
    hkl(th, tth, chi) -> (h, k, l)
        Return miller indicies for given IRIXS angles
    angles(h, k, l) -> (th, tth, chi)
        Return angles for given HKL
    find_hk_angles(list([h, k])) -> np.array([th, chi]):
        Return th and chi angles for list of HK values, given tth=90°
    update_B(a, b, c, alpha, beta, gamma)
        Update lattice parameters and B-matrix. UB-matrix recalculated.
    update_U(hkl0, hkl1, angles0, angles1)
        Update U-Matrix using two alignment reflections. UB-matrix recalculated.
    swap_alignment_refs()
        Swap hkl0 and hkl1 and recalculate the UB-matrix.
    """

    def __init__(self, cell, hkl0, hkl1, angles0, angles1=None, hkl1_offset=90, energy=2838.5):
        """
        Parameters
        ----------
        cell : list
            lattice parameters [a(Å), b(Å), c(Å), alpha(°), beta(°), gamma(°)]
        hkl0 : list
            first alignment reflection [h, k, l]
        hkl1 : list
            perp. second reflection [h, k, l]
        angles0 : list
            horiz angles for hkl0  [th0(°), tth0(°), chi0(°)]
        angles1 : list (optional)
            horiz angles for hkl1  [th1(°), tth1(°), chi1(°)]
        hkl1_offset : float
            if angles1 is None, assume hkl1 is hkl1_offset° away from hkl0
        energy : float
            x-ray energy (eV)
        """
        self.energy = energy
        self.wl = 12398 / energy  # eV -> Å
        self.hkl1_offset = hkl1_offset
        self.update_B(*cell)        
        self.update_U(hkl0, hkl1, angles0, angles1)

    def update_B(self, a, b, c, alpha, beta, gamma):
        """calculate B matrix from crystal lattice parameters"""

        a1, a2, a3 = a, b, c
        alpha1, alpha2, alpha3 = radians(alpha), radians(beta), radians(gamma)
        self.cell = [a1, a2, a3, alpha1, alpha2, alpha3]

        b1, b2, b3, beta1, beta2, beta3 = reciprocol_lattice(*self.cell)
        self.recip_cell = [b1, b2, b3, beta1, beta2, beta3]

        self.B = B_matrix(b1, b2, b3, beta2, beta3, a3, alpha1)
        self._update_UB()

    def update_U(self, hkl0, hkl1, angles0, angles1):
        """calculate the U Matrix using reflections hkl0 and hkl1"""

        # [th, tth, chi] degrees -> [mu, nu, chi] radians
        if angles1 is None:
            angles1 = deepcopy(angles0)
            angles1[0] += self.hkl1_offset
        angles0 = [radians(ang) for ang in angles0]
        angles1 = [radians(ang) for ang in angles1]
        self.orientation = [hkl0, hkl1, angles0, angles1]

        # HKL orientation vectors
        h1 = np.atleast_2d(np.array(hkl0)).T
        h2 = np.atleast_2d(np.array(hkl1)).T
        self._h1c = self.B.dot(h1)
        self._h2c = self.B.dot(h2)

        # Reflection vectors in phi frame
        self._u1p = q_phi(*angles0)
        self._u2p = q_phi(*angles1)

        self.U = U_matrix(self._h1c, self._h2c, self._u1p, self._u2p)
        self._update_UB()

    def swap_alignment_refs(self):
        """Swap hkl0 and hkl1 and recalculate the UB-matrix."""
        self._u1p, self._u2p = self._u2p, self._u1p
        self._h1c, self._h2c = self._h2c, self._h1c
        self.orientation = [
            self.orientation[1], self.orientation[0], self.orientation[3], self.orientation[2]
        ]
        self.U = U_matrix(self._h1c, self._h2c, self._u1p, self._u2p)
        self._update_UB()

    def _update_UB(self):
        """find the UB-matrix"""
        try:
            self.UB = self.U.dot(self.B)
        except AttributeError:
            pass

    def hkl(self, th, tth, chi):
        """return miller indices (h, k, l) for given IRIXS angles"""
        hkl = q_hkl(self.wl, self.UB, radians(th), radians(tth), radians(chi))
        return hkl

    def angles(self, h, k, l):
        """return IRIXS angles (th, tth, chi) for given reflection (h, k, l)"""
        x0 = [pi/4, pi/2, self.orientation[2][2]]  # init with th=45, tth=90, chi=chi0

        def fitfun(angles, h, k, l):
            hkl = q_hkl(self.wl, self.UB, angles[0], angles[1], angles[2])
            return hkl[0]-h, hkl[1]-k, hkl[2]-l

        result = least_squares(fitfun, x0, args=(h, k, l))
        return np.array([degrees(x) for x in result.x])

    def find_hk_angles(self, hk_list, printout=True):
        """find th and chi angles for list of HK values (ignoring L) with detector at 90°"""

        def fitfun(angles, h, k):
            hkl = q_hkl(self.wl, self.UB, angles[0], pi/2, angles[1])
            return hkl[0]-h, hkl[1]-k

        x0 = [pi/4, self.orientation[2][2]]  # init with th=45, chi=chi0

        angle_list = []
        for hk in hk_list:
            result = least_squares(fitfun, x0, args=hk)
            angles = [degrees(x) for x in result.x]
            angle_list.append(angles)
            if printout:
                hkl = q_hkl(self.wl, self.UB, result.x[0], pi/2, result.x[1])
                print(f"th{angles[0]: 9.4f}  chi{angles[1]: 9.4f}", end="  ")
                print(f"({hkl[0]: 6.3f} {hkl[1]: 6.3f} {hkl[2]: 6.3f})")

        return np.array(angle_list)


if __name__ == "__main__":

    # quick test based on Ca3Ru2O7
    cell = [5.368, 5.599, 19.35, 90, 90, 90]
    f = sixc(cell, [0,0,4], [1,0,0], [29.845, 53.6905, 2.0], energy=2838.5)

    # print hkl for values from grazing to normal
    for th in range(0, 100, 10):
        print(th, f.hkl(th, 90, 2.0).round(3))
    print()

    # swap hkl0 and hkl1
    f.swap_alignment_refs()

    # find angles for given hkl
    hkl = (1, 0.2, 6)
    angles = f.angles(*hkl)
    print(hkl, angles.round(3))
    print()

    # find angles for given hk and detector at 90° from grazing to normal
    hk_list = [(x, 0) for x in np.arange(-1.5, 1.6, 0.1)]
    angle_list = f.find_hk_angles(hk_list)

    # check it's working as expected
    assert(np.all(f.hkl(45, 90, 2).round(3)) == np.all([-0.091, 0, 6.257]))



""" reduction class for spectrograph """

import os
import shutil
import copy
import warnings
import numpy as np
import matplotlib.pyplot as plt

from glob import glob
from skimage import io
from matplotlib.offsetbox import AnchoredText

from .tools import load_fio, flatten, peak_fit, binning


def bias_correct_4output(rawimg):
    """
    for EFGH output
    divide matrix up into four quadrants and subtract median from each quadrant
    """
    r, h = rawimg.shape[1], rawimg.shape[0]
    nrows,ncols = h//2, r//2
    img = (rawimg
           .reshape((h//nrows,nrows,-1,ncols))
           .swapaxes(1,2)
           .reshape((-1, nrows, ncols)))
    for i,im in enumerate(img):
        cen = int(np.median(im))
        img[i] -= cen
    img = np.block([[img[0], img[1]], [img[2], img[3]]])
    return img


def load_tiff(filename, run_no, exp, datdir, localdir, correct=True):
    """ load greateyes detector tiff file
    correct - correct for bias of 4 quadrants when using EFGH output
    """
    tiffpath = '{dir}/{exp}_{run_no:05d}/greateyes/{filename}'
    path = tiffpath.format(dir=datdir, exp=exp, run_no=run_no, filename=filename)
    with warnings.catch_warnings():
        if localdir:
            path2 = tiffpath.format(dir=localdir, exp=exp, run_no=run_no, filename=filename)
            try:
                img = io.imread(path2)
            except (Warning, FileNotFoundError, OSError):
                try:
                    os.makedirs(os.path.dirname(path2), exist_ok=True)
                    shutil.copyfile(path, path2)
                    img = io.imread(path2)
                except IOError:
                    return
        else:
            try:
                img = io.imread(path)
            except (Warning, FileNotFoundError, OSError):
                return
    if correct:
        img = bias_correct_4output(img)
    return img


class spectrograph:

    def __init__(
        self,
        exp,
        detfac=0,
        threshold=15,
        cutoff=2000,
        roix=[0,2048],
        roih=2064,
        roic=2064//2,
        bias_correct=True,
        vmax_average=100,
        vmax_single=100,
        vmin=1,
        datdir_remote="/gpfs/current/raw",
        datdir_local="raw",
        savedir="processed",
    ):

        self.exp = exp
        self.detfac = detfac
        self.threshold = threshold
        self.cutoff = cutoff
        self.roix = roix
        self.roih = roih
        self.roic = roic
        self.bias_correct = bias_correct
        self.vmax_average = vmax_average
        self.vmax_single = vmax_single
        self.vmin = vmin
        self.datdir = datdir_remote
        self.localdir = datdir_local
        self.savedir = savedir

        os.makedirs(self.localdir, exist_ok=True)
        os.makedirs(self.savedir, exist_ok=True)

        self.runs = {}

    def extract(self, run_nos):

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]
        run_nos = flatten(run_nos)

        for n in run_nos:
            if n not in self.runs.keys():
                self.runs[n] = None

        for run_no in run_nos:
            if self.runs[run_no] and self.runs[run_no]["complete"]:
                continue

            fio_file = "{0}_{1:05d}.fio".format(self.exp, run_no)
            fio_remote = os.path.join(self.datdir, fio_file)
            fio_local = os.path.join(self.localdir, fio_file)

            if not os.path.isfile(fio_local):
                a = load_fio(run_no, self.exp, self.datdir)
                if a and a['complete']:
                    shutil.copyfile(fio_remote, fio_local)
            else:
                a = load_fio(run_no, self.exp, self.localdir)

            a["img"] = []
            print(f"{run_no}", end="")

            # since we don't have any control over image filename, sort by collection time
            # default to remote folder if it exists
            if os.path.exists(self.datdir):
                img_root = self.datdir
            else:
                img_root = self.localdir
            img_folder = os.path.join(img_root, f"{self.exp}_{run_no:05d}", "greateyes")
            filepaths = sorted(glob(img_folder+'*.tiff'), key=os.path.getctime)
            filenames = [os.path.basename(f) for f in filepaths]

        for i,f in enumerate(filenames):
            img = load_tiff(f, run_no, self.exp, self.datdir, self.localdir, self.bias_correct)
            if img is not None:
                img -= self.detfac
                img[~np.logical_and(img > self.threshold, img < self.cutoff)] = 0
                a['img'].append(img)
            else:
                break
        a['x'] = a['x'][:i+1]  # ensure length of x is same as number of images
        print(f'/{len(a["img"])}')

        self.runs[run_no] = a

    def transform(self, run_nos, ysca=1, fit=True):

        self.extract(run_nos)
        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        for run_no in run_nos:

            # sum up images if given a list of run_nos
            if isinstance(run_no, (list, tuple)):
                a = self.runs[run_no[0]]
                img = np.atleast_3d(np.array(a['img'])) / len(run_no)
                for r in run_no[1:]:
                    b = self.runs[r]
                    img = img + np.atleast_3d(np.array(b['img'])) / len(run_no)
            # single image
            else:
                a = self.runs[run_no]
                img = np.atleast_3d(np.array(a['img']))
                img = img * ysca

            x = a['x']
            if len(x) != len(img):
                x = x[:img.shape[0]]

            y, roi, imgr = [], [], []
            rx, ry, imgx, imgy = [], [], [], []

            for im, xi in zip(img, x):
                try:
                    # centre of roi as a lambda function
                    rc = self.roic(xi)
                except TypeError:
                    # fixed value
                    rc = self.roic
                roiy = rc - (self.roih//2),rc+(self.roih//2)
                roi.append([self.roix[0], self.roix[1], roiy[0], roiy[1]])

                ri = im[roiy[0]:roiy[1], self.roix[0]:self.roix[1]]
                imgr.append(ri)
                y.append(np.nansum(ri))

                rx.append(np.arange(roiy[0], roiy[1]))
                ry.append(np.arange(self.roix[0], self.roix[1]))

                imgx.append(np.nansum(ri, axis=1))
                imgy.append(np.nansum(ri, axis=0))

            y = np.array(y)

            r1 = min(i[0] for i in roi)
            r2 = max(i[1] for i in roi)
            r3 = min(i[2] for i in roi)
            r4 = max(i[3] for i in roi)
            x_totx = np.arange(r3, r4)
            x_toty = np.arange(r1, r2)

            imgroi = np.atleast_3d(np.array(imgr))
            tot = np.nansum(imgroi, axis=0) / img.shape[0]
            totx = np.nansum(tot, axis=1)
            toty = np.nansum(tot, axis=0)

            if fit:
                try:
                    xfx, yfx, px = peak_fit(x_totx, totx)
                    xfy, yfy, py = peak_fit(x_toty, toty)
                    txtx = f'fwhm: {px[1]*2:.4f}\ncen: {px[2]:.1f}\namp: {px[0]:.1f}\nfra: {px[3]:.1f}'
                    txty = f'fwhm: {py[1]*2:.4f}\ncen: {py[2]:.1f}\namp: {py[0]:.1f}\nfra: {py[3]:.1f}'
                except:
                    fit = False

            a["x"], a["y"] = x, y
            a["roi"], a["imgr"] = roi, imgroi
            a["rx"], a["ry"], a["imgx"], a["imgy"] = rx, ry, imgx, imgy
            a["extent"] = [r1, r2, r3, r4]
            a["tot"], a["totx"], a["toty"] = tot, totx, toty
            a["x_totx"] = x_totx
            a["x_toty"] = x_toty
            a["fitted"] = fit
            if fit:
                a["xfx"], a["yfx"], a["px"], a["txtx"] = xfx, yfx, px, txtx
                a["xfy"], a["yfy"], a["py"], a["txty"] = xfy, yfy, py, txty

    def detector(self, run_no):

        self.extract(run_no)

        if "x" not in self.runs[run_no]:
            self.transform(run_no)
        a = self.runs[run_no]

        fig = plt.figure(figsize=(16.5, 8.5))
        gs0 = fig.add_gridspec(
            1, 2, left=0.05, right=0.99, top=0.99, bottom=0.04, width_ratios=[1, 1.35]
        )

        # averaged total

        gs1 = gs0[0].subgridspec(
            3, 2, height_ratios=[1, 0.2, 0.4], width_ratios=[0.2, 1], hspace=0.1, wspace=0.1
        )
        ax1 = fig.add_subplot(gs1[0,1])
        ax2 = fig.add_subplot(gs1[0,0])
        ax3 = fig.add_subplot(gs1[1,1])
        ax4 = fig.add_subplot(gs1[2,:])

        fig.canvas.set_window_title(f'#{run_no}')

        ax1.imshow(
            a["tot"],
            origin='lower',
            interpolation='hanning',
            vmin=self.vmin,
            vmax=self.vmax_average,
            cmap=plt.get_cmap('bone_r'),
            extent=a["extent"],
            aspect='auto'
        )

        ax2.plot(a["totx"], a["x_totx"], color='k', lw=0.75)
        ax3.plot(a["x_toty"], a["toty"], color='k', lw=0.75)

        if a["fitted"]:
            ax2.plot(a["yfx"], a["xfx"], color='k',lw=0.5)
            ax3.plot(a["xfy"], a["yfy"], color='k',lw=0.5)
            ax2.add_artist(AnchoredText(a["txtx"], loc=1, frameon=False, pad=1))
            ax3.add_artist(AnchoredText(a["txty"], loc=2, frameon=False, pad=1))

        r1, r2, r3, r4 = a["extent"]
        ax2.set_ylim(r3, r4)
        ax3.set_xlim(r1, r2)

        # interactive singles

        img, roi, x = a["img"], a["roi"], a["x"]
        imgx, imgy, rx, ry = a["imgx"], a["imgy"], a["rx"], a["ry"]

        gs2 = gs0[1].subgridspec(
            2, 2, height_ratios=[1, 0.2], width_ratios=[0.2, 1], hspace=0.05, wspace=0.05
        )
        ax1 = fig.add_subplot(gs2[0,1])
        ax2 = fig.add_subplot(gs2[0,0])
        ax3 = fig.add_subplot(gs2[1,1])

        ax4.plot(a["x"], a["y"], color='k', lw=0.75)
        ax4.set_xlabel(a['auto'])
        ax4.set_ylabel('summed intensity')

        i = {'no':0}
        im = img[0]
        r = roi[0]

        cmap = copy.copy(plt.cmap.winter)
        cmap.set_under('w')
        cmap.set_over('#F012BE')

        i['a'] = ax1.imshow(
            im, origin='lower', vmin=self.vmin, vmax=self.vmax_single, cmap=cmap, aspect='auto'
        )
        i['b'], = ax2.plot(imgx[0], rx[0], color='k',lw=0.75)
        i['c'], = ax3.plot(ry[0], imgy[0], color='k',lw=0.75)
        i['v'] = ax4.axvline(x[0],color='r',lw=0.5)
        i['t'] = ax4.text(0.05, 0.9, x[0], transform=ax4.transAxes)

        ax1.set_xlim(r[0], r[1])
        ax1.set_ylim(r[2], r[3])
        ax2.set_ylim(r[2], r[3])
        ax3.set_xlim(r[0], r[1])

        def do_plot(i):
            im = img[i['no']]
            imx = imgx[i['no']]
            imy = imgy[i['no']]
            irx = rx[i['no']]
            iry = ry[i['no']]
            r = roi[i['no']]
            i['a'].set_data(im)
            i['b'].set_xdata(imx)
            i['b'].set_ydata(irx)
            i['c'].set_xdata(iry)
            i['c'].set_ydata(imy)
            i['v'].set_xdata(x[i['no']])
            i['t'].set_text(x[i['no']])
            ax2.relim()
            ax3.relim()
            ax2.autoscale(axis='x')
            ax3.autoscale(axis='y')
            plt.draw()

        def press(event):
            if event.key == 'left':
                if i['no'] > 0:
                    i['no'] -= 1
                    do_plot(i)
            elif event.key == 'right':
                if i['no'] < img.shape[0]-1:
                    i['no'] += 1
                    do_plot(i)

        def click(event):
            if not event.dblclick and event.button == 1:
                if event.inaxes in [ax4]:
                    i['no'] = np.abs(x-event.xdata).argmin()
                    do_plot(i)

        fig.canvas.mpl_connect('button_press_event', click)
        fig.canvas.mpl_connect('key_press_event', press)

    def condition(
        self,
        run_nos,
        bins=None,
        oneshot_x=False,
        oneshot_y=False,
        oneshot_no=None,
        fit=False,
        xsca=1,
        x0=0
    ):

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        self.transform(run_nos)

        for run_no in run_nos:
            if "x" not in self.runs[run_no]:
                self.transform(run_no)
            a = self.runs[run_no]

            if oneshot_x:
                if oneshot_no:
                    x = a['rx'][oneshot_no]
                    y = a['imgx'][oneshot_no]
                else:
                    x = a['x_totx']
                    y = a['totx']
                a["cond_type"] = "oneshot_x"
            elif oneshot_y:
                if oneshot_no:
                    x = a['ry'][oneshot_no]
                    y = a['imgy'][oneshot_no]
                else:
                    x = a['x_toty']
                    y = a['toty']
                a["cond_type"] = "oneshot_y"
            else:
                x, y = a['x'], a['y']
                a["cond_type"] = a["auto"]

            if bins:
                x, y, _ = binning(x, y, bins)

            x = (x - x0) * xsca

        a['cond_x'], a['cond_y'] = x, y

        if fit:
            try:
                xf, yf, p = peak_fit(x, y)
                r = '#{0}  '.format(a['no'])
                r += 'cen:{0:.4f}   '.format(p[2])
                r += 'amp:{0:.2f}   '.format(p[0])
                r += 'sig:{0:.4f}   '.format(p[1])
                r += 'fwhm:{0:.4f}  '.format(p[1]*2)
                r += 'fra:{0:.1f}   '.format(p[3])
                a['cond_xf'], a['cond_yf'], a['cond_p'], a['cond_r'] = xf, yf, p, r
                print(r)
            except:
                pass

        header = f"run: {a['no']}\nexp: {self.exp}\nbins: {bins}\n"
        if oneshot_x:
            header += "      y-pixel       intensity"
        elif oneshot_y:
            header += "      x-pixel       intensity"
        else:
            header += f"{a['auto']}    Intensity"
        np.savetxt(
            os.path.join(self.savedir, f"{self.exp}_{a['no']}_b{bins}.dat"),
            np.array([x, y]).T,
            fmt="% .8e",
            header=header
        )

    def plot(
        self,
        run_nos,
        ax=None,
        motor=None,
        norm=False,
        fit=False,
        ystep=0,
        plot_trend=False,
        **kwargs
    ):

        if not isinstance(run_nos, (list, tuple, range)):
            run_nos = [run_nos]

        if not ax:
            if plot_trend:
                fig = plt.figure(figsize=(7, 9), constrained_layout=True)
                gs = fig.add_gridspec(2, 3, height_ratios=[0.8, 1.0])
                axes = [
                    fig.add_subplot(gs[0,:]),
                    fig.add_subplot(gs[1,0]),
                    fig.add_subplot(gs[1,1]),
                    fig.add_subplot(gs[1,2])
                ]
            else:
                _, ax = plt.subplots(constrained_layout=True)

        if plot_trend:
            ax = axes[0]

        trend_x, trend_x0, trend_I, trend_fw = [], [], [], []
        for i, run_no in enumerate(run_nos):
            a = self.runs[run_no]
            x, y = a['cond_x'], a['cond_y']
            if "cond_xf" in a and fit:
                xf, yf, p = a['cond_xf'], a['cond_yf'], a['cond_p']
                if norm:
                    yf = yf/np.max(y)
            else:
                xf = None
            if norm:
                y = y/np.max(y)

            lab = f"#{run_no}"
            if motor:
                lab += f" {motor}: {a[motor]:.2f}"

            if fit and xf:
                trend_x.append(run_no)
                trend_x0.append(p[2])
                trend_fw.append(p[1]*2)
                trend_I.append(p[0])

            l, = ax.plot(x, y+i*ystep, label=lab, **kwargs)
            if fit and xf:
                ax.plot(xf, yf+i*ystep, color=l.get_color(), lw=0.5)

        ax.set_ylabel('Intensity')

        if a["cond_type"] == "oneshot_x":
            ax.set_xlabel('y-pixel')
        elif a["cond_type"] == "oneshot_y":
            ax.set_xlabel('x-pixel')
        else:
            ax.set_xlabel(a["cond_type"])

        if plot_trend:
            axes[1].plot(trend_x, trend_x0, '.-')
            axes[2].plot(trend_x, trend_fw, '.-')
            axes[3].plot(trend_x, trend_I, '.-')
            axes[1].set_ylabel('Center')
            axes[2].set_ylabel('FWHM')
            axes[3].set_ylabel('Amplitude')
            axes[3].set_ylim(0)
            for axi in axes:
                axi.set_xlabel("run number")

        ax.legend(fontsize='small')

    def track_signal(self, run_no, fit=True, bins=None, maxsig=1200, vert=False, title='no',
                     plot=True, plot_trend=False, ystep=0):

        a = self.runs[run_no]

        rx, ry, imgx, imgy = a['rx'], a['ry'], a['imgx'], a['imgy']

        if vert:
            r = rx
            im = imgx
        else:
            r = ry
            im = imgy

        txt_title = f'#{a["no"]}'
        if title != 'no':
            txt_title += f' {title}: {a[title]}'

        if plot:
            fig, ax = plt.subplots(figsize=(4,9), constrained_layout=True)
            ax.text(0.04, 0.98, txt_title, transform=ax.transAxes)

        trend_x0, trend_I, trend_fw = [], [], []
        for i, (x, y, xpos) in enumerate(zip(r, im, a['x'])):

            if bins:
                x, y, _ = binning(x, y, bins)

            if plot:
                l, = ax.plot(x, y+i*ystep, lw=0.75)
                ax.text(x[0], y[0]+i*ystep, f'{xpos:.4f}', fontsize='small')

            if fit:
                try:
                    xf, yf, p = peak_fit(x, y)
                    if plot:
                        ax.plot(xf, yf+i*ystep, color=l.get_color(), lw=0.5)

                    if (p[1] < maxsig) and (p[2] > 0):
                        trend_I.append(p[0])
                        trend_x0.append(p[2])
                        trend_fw.append(p[1]*2)
                    else:
                        trend_I.append(np.nan)
                        trend_x0.append(np.nan)
                        trend_fw.append(np.nan)    
                except:
                    trend_I.append(np.nan)
                    trend_x0.append(np.nan)
                    trend_fw.append(np.nan)
            else:
                com = np.sum(x*y)/np.sum(y)
                height = np.max(y) - np.min(y)
                half = x[np.abs(y-(height/2+np.min(y))).argmin()]
                wid = np.abs(half-com)*2
                trend_I.append(np.sum(y))
                trend_x0.append(com)
                trend_fw.append(wid)

        trend_x = a['x']

        if plot_trend:
            fig,ax = plt.subplots(1,3,constrained_layout=True,figsize=(10,4))

            ax[0].plot(trend_x, trend_I)
            ax[1].plot(trend_x, trend_x0)
            ax[2].plot(trend_x, trend_fw)

            ax[0].set_ylabel('intensity')
            ax[1].set_ylabel('centre')
            ax[2].set_ylabel('fwhm')

            ax[0].text(0.05, 0.95, txt_title, transform=ax[0].transAxes)

        # table = np.array([range(len(trend_x)),trend_x, trend_I, trend_x0, trend_fw]).T
        # print(txt_title)
        # print(tabulate(table,headers=['no', a['auto'], 'intensity', 'centre', 'fwhm']))
        # print()

        return np.array([trend_x, trend_I, trend_x0, trend_fw])



from setuptools import setup

with open('README.md') as readme_file:
    README = readme_file.read()

setup(
    name='IRIXS',
    version='0.6.6',
    description='Routines for the IRIXS spectrometer',
    long_description=README,
    long_description_content_type="text/markdown",
    url='http://github.com/djlns/irixs',
    author='Joel Bertinshaw',
    author_email='djlns@posteo.net',
    license='GNU GPLv3',
    install_requires=[
        'numpy>=1.18',
        'matplotlib>=3.1',
        'scipy>=1.4',
        'scikit-image>=0.16',
        'PyQt5>=5.9'
    ],
    zip_safe=False,
    packages=['IRIXS'],
    entry_points={
        'console_scripts': [
            'p01plot=IRIXS.p01plot:main',
            'irixs_oneshot=IRIXS.oneshot:main'],
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Image Processing',
        'Topic :: Scientific/Engineering :: Visualization'
    ]
)
